import inspect
import sys
from dataclasses import dataclass
from functools import singledispatch
from importlib import import_module
from typing import List, Optional, Union

from doctorr.autodoc.parse_config import (
    ClassEntry,
    FunctionEntry,
    MethodEntry,
    PageConfig,
    PropertyEntry,
)


@dataclass
class InspectedProperty:
    name: str
    docstring: Optional[str]


@dataclass
class InspectedMethod:
    name: str
    signature: str
    docstring: Optional[str]


@dataclass
class InspectedFunction:
    name: str
    signature: str
    docstring: Optional[str]


@dataclass
class InspectedClass:
    name: str
    init_signature: Optional[str]
    docstring: Optional[str]
    members: List[Union[InspectedMethod, InspectedProperty]]


@dataclass
class InspectedPage:
    title: str
    output_name: str
    module_source: str
    user_import_path: str
    module_docstring: Optional[str]
    content: List[Union[InspectedClass, InspectedFunction]]


@singledispatch
def inspect_entry(entry, parent):
    raise NotImplementedError(f"Cannot handle {type(entry)}")


@inspect_entry.register(FunctionEntry)
def _(entry: FunctionEntry, parent):
    func = getattr(parent, entry.name, None)
    if func and callable(func):
        signature = str(inspect.signature(func))
        return InspectedFunction(
            name=func.__name__, signature=signature, docstring=func.__doc__
        )


@inspect_entry.register(ClassEntry)
def _(entry: ClassEntry, parent):
    cls = getattr(parent, entry.name, None)
    if not cls:
        raise AttributeError(entry.name)

    members = []
    for member_config in entry.members:
        extracted_member = inspect_entry(member_config, cls)
        if extracted_member:
            members.append(extracted_member)

    init_signature = None
    if hasattr(cls, "__init__"):
        init_signature = (
            str(inspect.signature(cls.__init__))
            .replace("(self, ", "(")
            .replace("(self)", "()")
        )

    return InspectedClass(
        name=cls.__name__,
        init_signature=init_signature,
        docstring=cls.__doc__,
        members=members,
    )


@inspect_entry.register(MethodEntry)
def _(entry: MethodEntry, parent):
    method = getattr(parent, entry.name, None)
    if method and callable(method):
        signature = str(inspect.signature(method))
        return InspectedMethod(
            name=method.__name__, signature=signature, docstring=method.__doc__
        )


@inspect_entry.register(PropertyEntry)
def _(entry: PropertyEntry, parent):
    prop = getattr(parent, entry.name, None)
    if isinstance(prop, property):
        return InspectedProperty(name=prop.fget.__name__, docstring=prop.__doc__)


def import_and_inspect(page_config: PageConfig, watch: bool) -> InspectedPage:
    module_name = page_config.module_source
    # hot module reload
    if watch and module_name in sys.modules:
        del sys.modules[module_name]

    module = import_module(module_name)
    module_docstring = module.__doc__ if page_config.include_module_docstring else None

    content_list = []
    for item in page_config.content:
        content_item = inspect_entry(item, module)
        if content_item:
            content_list.append(content_item)

    return InspectedPage(
        title=page_config.title,
        output_name=page_config.output_name,
        module_source=page_config.module_source,
        user_import_path=page_config.user_import_path,
        module_docstring=module_docstring,
        content=content_list,
    )



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

from doctorr.parse_config import _parse_yaml, _read_file


@dataclass
class PropertyEntry:
    name: str


@dataclass
class MethodEntry:
    name: str


@dataclass
class FunctionEntry:
    name: str


@dataclass
class ClassEntry:
    name: str
    members: List[Union[MethodEntry, PropertyEntry]]


@dataclass
class PageConfig:
    title: str
    output_name: str
    module_source: str
    user_import_path: str
    content: List[Union[ClassEntry, FunctionEntry]]
    include_module_docstring: Optional[bool] = False


def _create_page_config(cfg: dict) -> PageConfig:
    content_list = []
    for item in cfg.pop("content"):
        if "class" in item:
            class_name = item["class"]
            members = []
            for member in item["members"]:
                if "method" in member:
                    members.append(MethodEntry(name=member["method"]))
                elif "property" in member:
                    members.append(PropertyEntry(name=member["property"]))
            content_list.append(ClassEntry(name=class_name, members=members))
        elif "function" in item:
            content_list.append(FunctionEntry(name=item["function"]))

    if "user_import_path" not in cfg:
        cfg["user_import_path"] = cfg["module_source"]

    return PageConfig(content=content_list, **cfg)


def main(config_path: str) -> List[PageConfig]:
    file_content = _read_file(config_path)
    config = _parse_yaml(file_content)
    return config["autodoc"]


def parse_page_configs(config):
    return [_create_page_config(page_cfg) for page_cfg in config["pages"]]


def parse_watch_config(config):
    return [Path(p) for p in config["watch"]]


def parse_output_dir(config):
    return Path(config["output_dir"])



from enum import Enum
from shutil import rmtree

from doctorr.autocli.inspection import get_parser
from doctorr.autocli.parse_argparse import parse_parser
from doctorr.autocli.parse_config import main as read_config
from doctorr.autocli.templates import render_cli
from doctorr.include import run_include
from doctorr.parse_config import parse_include_config
from doctorr.templates import TemplateRenderer
from doctorr.watchmode import main as watchmode


class AutoCliDefault(Enum):
    CONFIG_PATH = "./doctorr.yml"
    THEME = "md_basic"


def run_autocli(config_path, theme, output_dir, watch, **_):
    config = read_config(config_path)
    _run_autocli(config_path, theme, output_dir)
    if watch and config.watch:
        watchmode(lambda: _run_autocli(config_path, theme, output_dir), config)


def _run_autocli(config_path, theme, output_dir):
    config = read_config(config_path)
    parser = get_parser(config)
    renderer = TemplateRenderer(theme)
    text = generate_cli_documentation(parser, renderer)

    output_dir = output_dir or config.output_dir
    assert output_dir, "Please provide through cli or through the yaml"

    rmtree(output_dir, ignore_errors=True)
    output_dir.mkdir(parents=True, exist_ok=True)
    (output_dir / config.output_name).write_text(text)

    include_configs = parse_include_config(config)
    run_include(include_configs)


def generate_cli_documentation(parser, renderer):
    parsed_data = parse_parser(parser)
    markdown_doc = render_cli(parsed_data, renderer)
    return markdown_doc



import time
from threading import Timer

from watchdog.events import FileModifiedEvent, FileSystemEventHandler
from watchdog.observers import Observer

from doctorr.utils import _get


def debounce(wait):
    """Decorator that will postpone a functions
    execution until after wait seconds
    have elapsed since the last time it was invoked."""

    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)

            try:
                debounced.t.cancel()
            except AttributeError:
                pass
            debounced.t = Timer(wait, call_it)
            debounced.t.start()

        return debounced

    return decorator


def main(func, config):
    @debounce(0.1)
    def build_doc():
        # watchdog sees more events than actually happen,
        # this makes sure it runs once
        print("rebuilding...")
        func()

    class _Watcher(FileSystemEventHandler):
        def on_modified(self, event):
            if (
                "__pycache__" in event.src_path
                or not isinstance(event, FileModifiedEvent)
                or event.src_path.split(".")[-1] not in ["py", "yaml", "yml", "jinja2"]
            ):
                return
            try:
                build_doc()
            except KeyboardInterrupt as e:
                raise e
            except Exception as e:
                # todo: improve but dont quit on syntax error
                print(e)

    event_handler = _Watcher()
    observer = Observer()
    for path in _get("watch", config):
        print("watching", path)
        observer.schedule(event_handler, path=path, recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



import argparse
import os


def parse_parser(parser, data=None, **kwargs):
    """Enhanced parser function to handle complex argparse structures."""
    if data is None:
        data = {
            "name": "",
            "usage": parser.format_usage().strip(),
            "bare_usage": _format_usage(parser),
            "prog": parser.prog,
        }

    data.setdefault("description", getattr(parser, "description", None))
    data.setdefault("epilog", getattr(parser, "epilog", None))

    arguments = []
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            subparsers = []

            for i, (choice, subparser) in enumerate(action.choices.items()):
                subparser_data = parse_parser(subparser)
                subparser_data["name"] = choice
                subparser_data["subcommand_usage"] = _format_usage(subparser)
                subparsers.append(subparser_data)
                subparser_data["description"] = getattr(subparser, "description", None)
                subparser_data["help"] = action._choices_actions[i].help
            data["subcommands"] = subparsers
        else:
            argument_data = {
                "name": action.dest,
                "help": action.help,
                "default": action.default,
                "choices": action.choices,
            }
            arguments.append(argument_data)

    data["arguments"] = arguments
    return data


def _format_usage(parser):
    usage_lines = _get_usage(parser).strip().split("\n")
    indent = len("usage: ")
    usage_lines = [line[indent:] for line in usage_lines]
    return "\n".join(usage_lines)


def _set_terminal_column_width(col_width):
    def decorator(func):
        def wrapper(*args, **kwargs):
            # Store the original COLUMNS value
            original_columns = os.environ.get("COLUMNS")
            # Set a custom terminal width
            os.environ["COLUMNS"] = "80"
            # decorated function call
            ret = func(*args, **kwargs)
            # Restore the original COLUMNS value
            if original_columns:
                os.environ["COLUMNS"] = original_columns
            else:
                os.environ.pop("COLUMNS", None)
            return ret

        return wrapper

    return decorator


@_set_terminal_column_width(80)
def _get_usage(parser):
    return parser.format_usage()  # takes in consideration terminal's width



from enum import Enum
from shutil import rmtree

from doctorr.autodoc.docstring import process_page
from doctorr.autodoc.inspection import import_and_inspect
from doctorr.autodoc.parse_config import main as parse_config
from doctorr.autodoc.parse_config import parse_output_dir, parse_page_configs
from doctorr.autodoc.templates import render_page
from doctorr.include import run_include
from doctorr.parse_config import parse_include_config
from doctorr.templates import TemplateRenderer
from doctorr.watchmode import main as watchmode


class AutoDocDefault(Enum):
    CONFIG_PATH = "./doctorr.yml"
    THEME = "md_basic"


def run_autodoc(config_path, theme, output_dir, watch, **_):
    config = parse_config(config_path)
    _run_autodoc(config_path, theme, output_dir, watch)
    if watch and "watch" in config:
        watchmode(lambda: _run_autodoc(config_path, theme, output_dir, watch), config)


def _run_autodoc(config_path, theme, output_dir, watch):
    config = parse_config(config_path)
    page_configs = parse_page_configs(config)
    inspected_pages = [import_and_inspect(page, watch) for page in page_configs]
    processed_pages = [process_page(page) for page in inspected_pages]
    renderer = TemplateRenderer(theme)
    rendered_pages = [render_page(page, renderer) for page in processed_pages]

    output_dir = output_dir or parse_output_dir(config)

    rmtree(output_dir, ignore_errors=True)
    output_dir.mkdir(parents=True, exist_ok=True)
    for page, text in zip(processed_pages, rendered_pages):
        (output_dir / page.output_name).write_text(text)

    include_configs = parse_include_config(config)
    run_include(include_configs)



from dataclasses import dataclass
from pathlib import Path

import yaml

from doctorr.utils import _get


def _read_file(path) -> str:
    return Path(path).read_text()


def _parse_yaml(text) -> str:
    return yaml.load(text, Loader=yaml.FullLoader)


@dataclass
class IncludeConfig:
    original: Path
    destination: Path

    def __post_init__(self):
        self.original = Path(self.original)
        self.destination = Path(self.destination)


def parse_include_config(config):
    return [IncludeConfig(**c) for c in _get("include", config, [])]


# todo: watch parsing should probably be there too.



import shutil

from doctorr.parse_config import IncludeConfig


def run_include(include_configs: list[IncludeConfig]):
    for config in include_configs:
        # Ensure the destination directory exists
        config.destination.parent.mkdir(parents=True, exist_ok=True)

        # Check if it's a file or directory
        if config.original.is_file():
            shutil.copy2(config.original, config.destination)
        elif config.original.is_dir():
            shutil.copytree(config.original, config.destination)



from pathlib import Path

from jinja2 import Environment, PackageLoader, select_autoescape


class TemplateRenderer:
    def __init__(
        self,
        theme_name,
        *,
        theme_dir=Path("templates/themes"),
        package_name="doctorr",
    ):
        self.env = Environment(
            loader=PackageLoader(
                package_name=package_name, package_path=theme_dir / theme_name
            ),
            autoescape=select_autoescape(
                default_for_string=False,
                default=False,
            ),
            trim_blocks=True,
        )

    def render(self, template_name, context):
        template = self.env.get_template(template_name)
        return template.render(context)



import contextvars
from functools import singledispatch

from doctorr.autodoc.docstring import (
    ProcessedClass,
    ProcessedFunction,
    ProcessedMethod,
    ProcessedPage,
    ProcessedProperty,
)

page_context_var = contextvars.ContextVar('page')

@singledispatch
def _render_element(processed_element, template_renderer) -> str:
    raise ValueError(f"Unsupported ProcessedElement type: {type(processed_element)}")


def format_user_import_path(user_import_path):
    if user_import_path:
        user_import_path += "."
    return user_import_path


def element_to_template_dict(processed_element):
    _page_context = page_context_var.get()
    user_import_path = format_user_import_path(
        _page_context.user_import_path 
    )
    return {
        "name": processed_element.name,
        "signature": getattr(processed_element, "signature", ""),
        "docstring": processed_element.docstring,
        "user_import_path": user_import_path,
    }


@_render_element.register(ProcessedClass)
def _(processed_element, renderer) -> str:
    context = element_to_template_dict(processed_element)
    template_name = "class.jinja2"
    rendered_members = [
        _render_element(member, renderer) for member in processed_element.members
    ]
    context["rendered_members"] = "".join(rendered_members)
    return renderer.render(template_name, context)


@_render_element.register(ProcessedFunction)
def _(processed_element, renderer) -> str:
    context = element_to_template_dict(processed_element)
    template_name = "function.jinja2"
    return renderer.render(template_name, context)


@_render_element.register(ProcessedMethod)
def _(processed_element, renderer) -> str:
    context = element_to_template_dict(processed_element)
    template_name = "method.jinja2"
    return renderer.render(template_name, context)


@_render_element.register(ProcessedProperty)
def _(processed_element, renderer) -> str:
    context = element_to_template_dict(processed_element)
    template_name = "property.jinja2"
    return renderer.render(template_name, context)


@_render_element.register(ProcessedPage)
def _(processed_element, renderer) -> str:
    context = element_to_template_dict(processed_element)
    template_name = "page.jinja2"
    rendered_content = [
        _render_element(piece_of_content, renderer)
        for piece_of_content in processed_element.content
    ]
    context["rendered_content"] = "".join(rendered_content)
    return renderer.render(template_name, context)


def render_page(page: ProcessedPage, renderer):
    page_context_var.set(page)
    return _render_element(page, renderer)



import argparse
from pathlib import Path

from doctorr.autocli import AutoCliDefault, run_autocli
from doctorr.autodoc import AutoDocDefault, run_autodoc


def main():
    """CLI entrypoint"""
    args = _parse_command()
    command = TARGET_COMMANDS[args.target]
    return command(args)


def _parse_command():
    args = _get_parser().parse_args()
    return args


def autodoc_command(cli_args):
    run_autodoc(**vars(cli_args))


def autocli_command(cli_args):
    run_autocli(**vars(cli_args))


TARGET_COMMANDS = {"autodoc": autodoc_command, "autocli": autocli_command}


def _get_parser(*subcommands):
    parser = argparse.ArgumentParser(
        description="doctorr - next level autodoc tooling for python pros",
        prog="doctorr",
    )

    targets = parser.add_subparsers(title="targets", dest="target")

    autodoc = targets.add_parser(
        "autodoc", help="Generate documentation from docstrings automatically."
    )

    autodoc.add_argument(
        "-c",
        "--config_path",
        default=AutoDocDefault.CONFIG_PATH.value,
        type=Path,
        help="Path to the yaml config file for generating package documentation.",
    )

    autodoc.add_argument(
        "-t",
        "--theme",
        default=AutoDocDefault.THEME.value,
        help="Name of the default theme to use.",
    )

    autodoc.add_argument(
        "-o",
        "--output_dir",
        required=False,
        type=Path,
        help="Path to desired output directory for the generated files.",
    )

    autodoc.add_argument(
        "-w",
        "--watch",
        action="store_true",
        help="When the the content of the path changes, will rebuild the doc.",
    )

    autocli = targets.add_parser(
        "autocli", help="Generate documentation from cli parser automatically."
    )

    autocli.add_argument(
        "-c",
        "--config_path",
        default=AutoCliDefault.CONFIG_PATH.value,
        type=Path,
        help="Path to the yaml config file for generating CLI documentation.",
    )

    autocli.add_argument(
        "-t",
        "--theme",
        default=AutoCliDefault.THEME.value,
        help="Name of the default theme to use.",
    )

    autocli.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        required=False,
        help="Path to desired output directory for the generated files.",
    )

    autocli.add_argument(
        "-w",
        "--watch",
        action="store_true",
        help="When the the content of the path changes, will rebuild the doc.",
    )

    return parser



import sys


def get_parser(config):
    if config.module_source in sys.modules:
        del sys.modules[config.module_source]
    module = __import__(config.module_source, fromlist=[config.getter])
    getter = getattr(module, config.getter)
    parser = getter()
    return parser



if __name__ == "__main__":
    from doctorr.cli import _get_parser

    from .run import generate_cli_documentation

    parser = _get_parser()
    markdown = generate_cli_documentation(parser)
    print(markdown)



def _get(key, obj, default=None):
    # bad design:
    # config["include"] in autodoc
    # config.include in autocli
    # (same for "watch")
    # this works around this bad design
    if isinstance(obj, dict):
        return obj.get(key, default)
    return getattr(obj, key, default)



from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

from doctorr.include import IncludeConfig
from doctorr.parse_config import _parse_yaml, _read_file


@dataclass
class CLIConfig:
    module_source: str
    getter: str
    prog: str
    output_name: str
    output_dir: Optional[Path] = None
    watch: List[str] = field(default_factory=list)
    include: List[IncludeConfig] = field(default_factory=list)

    def __post_init__(self):
        self.output_dir = self.output_dir and Path(self.output_dir)


def main(config_path):
    file_content = _read_file(config_path)
    full_config = _parse_yaml(file_content)
    autocli_config = full_config["autocli"]
    return CLIConfig(**autocli_config)



import re
from dataclasses import dataclass, field, fields
from functools import singledispatch
from itertools import zip_longest
from textwrap import dedent
from typing import ClassVar, List, Optional, Union

from doctorr.autodoc.inspection import (
    InspectedClass,
    InspectedFunction,
    InspectedMethod,
    InspectedPage,
    InspectedProperty,
)

# note  docstring is a bad name for this module.
#       it does more than handling docstring


@dataclass
class Field:
    name: str
    type: Optional[str]
    text: str


@dataclass
class Docstring:
    description: str = ""
    attributes: List[Field] = field(default_factory=list)
    args: List[Field] = field(default_factory=list)
    returns: List[Field] = field(default_factory=list)
    yields: List[Field] = field(default_factory=list)
    raises: List[Field] = field(default_factory=list)

    @property
    def sections(self):
        return [
            {
                "title": Docstring.MAP_SECTION_NAME_TO_TITLE[f.name],
                "fields": getattr(self, f.name),
            }
            for f in fields(self)
            if f.name != "description"
        ]

    MAP_SECTION_NAME_TO_TITLE: ClassVar = {
        "attributes": "Attributes",
        "args": "Parameters",
        "returns": "Returns",
        "yields": "Yields",
        "raises": "Raises",
    }


@dataclass
class ProcessedFunction:
    name: str
    signature: str
    docstring: Docstring


@dataclass
class ProcessedMethod:
    name: str
    signature: str
    docstring: Docstring


@dataclass
class ProcessedProperty:
    name: str
    docstring: Docstring


@dataclass
class ProcessedClass:
    name: str
    docstring: Docstring
    members: List[Union[ProcessedMethod, ProcessedProperty]]
    signature: str = "()"


@dataclass
class ProcessedPage:
    title: str
    output_name: str
    module_source: str
    user_import_path: str
    docstring: Docstring
    content: List[Union[ProcessedClass, ProcessedFunction]]

    @property
    def name(self):
        return self.title


def parse_google_docstring(docstring: str) -> Docstring:
    if not docstring:
        return Docstring()  # empty docstring
    sections = _split_into_sections(docstring)
    description = _dedent_description(sections.get("Description", ""))
    args = _parse_fields_section(sections.get("Parameters", ""))
    returns = _parse_fields_section(sections.get("Returns", ""))
    yields = _parse_fields_section(sections.get("Yields", ""))
    raises = _parse_fields_section(sections.get("Raises", ""))
    attributes = _parse_fields_section(sections.get("Attributes", ""))
    return Docstring(description, attributes, args, returns, yields, raises)


def _dedent_description(description):
    """Remove common leading whitespace from every line in the given description.

    This function is designed to handle typical Python docstrings, where the first line
    of the docstring might be on the same line as the opening triple quotes, followed by
    a newline, and subsequent lines may have a common indentation level.
    """
    if not description:
        return ""
    # the first line of a docstring maybe indented differenly than the rest
    first_line, *rest = description.split("\n")
    first_line = first_line.strip()
    rest = dedent("\n".join(rest))
    return "\n".join([first_line, rest])


SECTION_REGEX = re.compile(
    r"^\s*(Args:|Arguments:|Parameters:|Returns:|Return:|Raises:|Yields:|"
    "Attributes:)",
    re.MULTILINE,
)

# This is used to map aliases toward the unique sections
SECTION_MAP = {
    "Args:": "Parameters",
    "Arguments:": "Parameters",
    "Parameters:": "Parameters",
    "Returns:": "Returns",
    "Return:": "Returns",
    "Raises:": "Raises",
    "Yields:": "Yields",
    "Attributes:": "Attributes",
    "Description": "Description",  # no : because it's a field we are creating
}


def _identify_function_intervals(docstring):
    """Identify where the sections starts and end as well as their title.

    Returns a list of dict with keys {"title", "start", "end"}.

    Warning: "start" and "end" are the indexes in the full docstring
    (not the line numbers)
    """
    matches = list(SECTION_REGEX.finditer(docstring))
    sections = [{"start": m.start(), "title": m.group(0).strip()} for m in matches]
    sections.insert(0, {"start": 0, "title": "Description"})
    sections = [
        {**section, "end": next_["start"]}
        for section, next_ in zip_longest(
            sections, sections[1:], fillvalue={"start": None}
        )
    ]
    return sections


def _check_for_duplicate_sections(sections):
    if len(set([SECTION_MAP[s["title"]] for s in sections])) < len(sections):
        raise ValueError("Duplicate section detected.")


def _split_into_sections(docstring: str) -> dict:
    sections = _identify_function_intervals(docstring)
    _check_for_duplicate_sections(sections)
    return {
        SECTION_MAP[section["title"]]: docstring[section["start"] : section["end"]]
        for section in sections
    }


_field_parens_regex = re.compile(r"\s*(\w+)\s*\(\s*(.+?)\s*\)")


def _parse_fields_section(section_text: str) -> List[Field]:
    fields = []
    lines = section_text.strip().split("\n")
    lines = lines[1:]  # skip section title

    for line in lines:
        if ":" in line:
            _name, _, _desc = line.partition(":")
            _name, _type, _desc = _name.strip(), "", _desc.strip()
            match = _field_parens_regex.match(_name)
            if match:
                _name = match.group(1)
                _type = match.group(2)
            fields.append(Field(name=_name, type=_type, text=_desc.strip()))
        elif fields:
            # This is a continuation of the description of the previous field
            fields[-1].text += " " + line.strip()

    return fields


@singledispatch
def process_inspected_item(item):
    raise NotImplementedError(f"Cannot handle {type(item)}")


@process_inspected_item.register(InspectedFunction)
def _(item: InspectedFunction):
    doc = parse_google_docstring(item.docstring)
    return ProcessedFunction(name=item.name, signature=item.signature, docstring=doc)


@process_inspected_item.register(InspectedClass)
def _(item: InspectedClass):
    class_doc = parse_google_docstring(item.docstring)
    processed_members = []
    for member in item.members:
        processed_member = process_inspected_item(member)
        if processed_member:
            processed_members.append(processed_member)
    return ProcessedClass(
        name=item.name,
        signature=item.init_signature,
        docstring=class_doc,
        members=processed_members,
    )


@process_inspected_item.register(InspectedMethod)
def _(item: InspectedMethod):
    doc = parse_google_docstring(item.docstring)
    return ProcessedMethod(name=item.name, signature=item.signature, docstring=doc)


@process_inspected_item.register(InspectedProperty)
def _(item: InspectedProperty):
    doc = parse_google_docstring(item.docstring)
    return ProcessedProperty(name=item.name, docstring=doc)


def process_page(inspected_page: InspectedPage) -> ProcessedPage:
    processed_content = [
        process_inspected_item(item) for item in inspected_page.content
    ]
    module_doc = parse_google_docstring(inspected_page.module_docstring)

    return ProcessedPage(
        title=inspected_page.title,
        output_name=inspected_page.output_name,
        module_source=inspected_page.module_source,
        user_import_path=inspected_page.user_import_path,
        docstring=module_doc,
        content=processed_content,
    )
